function UI_Rect(x, y, w, h) {
	var rect = Object.create(UI_Rect);
	rect.x = +x || 0;
	rect.y = +y || 0;
	rect.w = +w || 10;
	rect.h = +h || 10;
	return rect;
}

if ( !Array.prototype.forEach ) {
	Array.prototype.forEach = function( callback, thisArg ) {
		var T, k;
		if ( this == null ) {
			throw new TypeError( " this is null or not defined" );
		}
		// 1. Let O be the result of calling ToObject passing the |this| value as the argument.
		var O = Object(this);
		// 2. Let lenValue be the result of calling the Get internal method of O with the argument "length".
		// 3. Let len be ToUint32(lenValue).
		var len = O.length >>> 0; // Hack to convert O.length to a UInt32
		// 4. If IsCallable(callback) is false, throw a TypeError exception.
		// See: http://es5.github.com/#x9.11
		if ( {}.toString.call(callback) != "[object Function]" ) {
			throw new TypeError( callback + " is not a function" );
		}
		// 5. If thisArg was supplied, let T be thisArg; else let T be undefined.
		if ( thisArg ) {
			T = thisArg;
		}
		// 6. Let k be 0
		k = 0;

		// 7. Repeat, while k < len
		while( k < len ) {
			var kValue;
			// a. Let Pk be ToString(k).
			//   This is implicit for LHS operands of the in operator
			// b. Let kPresent be the result of calling the HasProperty internal method of O with argument Pk.
			//   This step can be combined with c
			// c. If kPresent is true, then
			if ( k in O ) {
				// i. Let kValue be the result of calling the Get internal method of O with argument Pk.
				kValue = O[ k ];
				// ii. Call the Call internal method of callback with T as the this value and
				// argument list containing kValue, k, and O.
				callback.call( T, kValue, k, O );
			}
			// d. Increase k by 1.
			k++;
		}
		// 8. return undefined
	};
}

if (Object.prototype.assign === undefined) {
	Object.prototype.assign = function(target, source) {
//		for (var source of sources) {
			var keys = Object.getOwnPropertyNames(source);
			for (var p in keys) {
				var key = keys[p];
				var desc = Object.getOwnPropertyDescriptor(source, key);
				if (desc && desc.enumerable) {
					target[key] = source[key];
				}
			}
//		}
	};
}

UI_Rect.isinside = function(x, y) {
	x = +x || 0;
	y = +y || 0;
	if (this.x <= x && x < this.x + this.w && this.y <= y && y < this.y + this.h) {
		return true;
	}
	else {
		return false;
	}
};
UI_Rect.prototype = {}

function UI_Style(param) {
	var style = Object.create(UI_Style);
	style.font	=					(param && param.font)					||	"Century, sans serif";
	style.fontsize =			(param && param.fontsize)			|| 12;
	style.textStyle =			(param && param.textStyle)		|| "black";
	style.strokeStyle =		(param && param.strokeStyle)	|| "black";
	style.lineWidth =			(param && param.lineWidth)		|| 0;
	style.fillStyle =			(param && param.fillStyle)		|| "transparent";
	style.align =					(param && param.align)				|| "left";
	style.valign =				(param && param.valign)				|| "middle";
	style.px =						(param && param.px)						|| 0;
	style.py =						(param && param.py)						|| 0;
	return style;
}
UI_Style.prototype = {};

function UI_Drawdata(type, param) {
	var drawdata =						Object.create(UI_Drawdata);
	drawdata.type =						(param && param.type) || "text";
	drawdata.id =							(param && param.id !== undefined) ? param.id : drawdata.type;
	drawdata.style =					UI_Style(param && param.style);
	drawdata.moverstyle =			(param && param.moverstyle) ? UI_Style(param.moverstyle) : drawdata.style;
	drawdata.offstyle =				(param && param.offstyle) ? UI_Style(param.offstyle) : drawdata.style;
	drawdata.moveroffstyle =	(param && param.moveroffstyle) ? UI_Style(param.moveroffstyle) : drawdata.offstyle;
	drawdata.x =							(param && param.x) || 0;
	drawdata.y = 							(param && param.y) || 0;
	drawdata.width =					(param && param.width) || 50;
	drawdata.height	=					(param && param.height) || 50;
	drawdata.position =				(param && param.position) || "auto";
	drawdata.dx =							(param && param.dx) || 0;
	drawdata.dy =							(param && param.dy) || 0;
	drawdata.transform =			(param && param.transform) || null;
	drawdata.param =					(param && param.param) || null;
	drawdata.prototype.update();
	drawdata.mouseover = false;
	drawdata.on	= true;
	return drawdata;
}

UI_Drawdata.prototype = {
	gettype: function() { return this.type; },
	update:	function() {
		this.rect =		UI_Rect(this.x, this.y, this.width, this.height);
	},
	drawbackground:	function(ui) {
		var ctx = ui.ctx;
		var style = (this.on) ? ((this.mouseover) ? this.moverstyle : this.style) : ((this.mouseover) ? this.moveroffstyle : this.offstyle);
		ctx.beginPath();
		ctx.rect(this.x, this.y, this.width, this.height);
		ctx.fillStyle = style.fillStyle;
		ctx.fill();
		if (style.lineWidth > 0) {
			ctx.beginPath();
			ctx.lineWidth = style.lineWidth;
			ctx.strokeStyle = style.strokeStyle;
			ctx.rect(this.x + style.lineWidth / 2, this.y + style.lineWidth / 2, this.width - style.lineWidth, this.height - style.lineWidth);
			ctx.strokeStyle = style.strokeStyle;
			ctx.stroke();
		}
	}
};

function UI_Drawdata_Text(param) {
	var ddata = Object.create(UI_Drawdata_Text.prototype);
	Object.assign(ddata, UI_Drawdata("text", param));
	ddata.text				= (param && param.text) || "";
	ddata.bindvar			= (param && param.bindvar) || null;
	return ddata;
}

UI_Drawdata_Text.prototype = {
	draw:	function(ui) {
		var ctx = ui.ctx;
		this.drawbackground(ui);
		var style = (this.on) ? ((this.mouseover) ? this.moverstyle : this.style) : ((this.mouseover) ? this.moveroffstyle : this.offstyle);
		ctx.fillStyle	= style.textStyle;
		ctx.textBaseline = "bottom";
		ctx.font			= style.fontsize + "px " + style.font;
		var text = this.text = (this.bindvar) ? eval(this.bindvar) : this.text;
		var twidth = ctx.measureText(text).width;
		var x = this.x + style.lineWidth + style.px;
		var y = this.y + style.lineWidth + style.fontsize + style.py;
		if (style.align === "center") {
			x = this.x + (this.width - twidth) / 2;
		}
		else if (style.align === "right") {
			x = this.x + this.width - twidth - style.lineWidth - style.px;
		}
		if (style.valign === "middle") {
			y = this.y + (style.fontsize + this.height) / 2;
		}
		else if (style.valign === "bottom") {
			y = this.y + this.height - style.lineWidth - style.py;
		}			
		ctx.fillText(text, x, y);
	}
};

Object.setPrototypeOf(UI_Drawdata_Text.prototype, UI_Drawdata.prototype);

function UI_Drawdata_Image(param) {
	var ddata = Object.create(UI_Drawdata_Image.prototype);
	Object.assign(ddata, UI_Drawdata("image", param));
	if (param && param.image) {
		ddata.image = param.image;
	}
	else {
		ddata.image	=				new Image();
		ddata.image.src =		(param && param.src) || null;
	}
	ddata.image.ddata = ddata;
	ddata.image.param = param;
	ddata.image.onload = function(param) {
		this.ddata.imagewidth =		(this.param && this.param.imagewidth) || this.ddata.image.width;
		this.ddata.imageheight =	(this.param && this.param.imageheight) || this.ddata.image.height;
	}
	ddata.imagewidth =	(param && param.imagewidth) || ddata.image.width;
	ddata.imageheight =	(param && param.imageheight) || ddata.image.height;
	ddata.imagesrcx =		(param && param.imagesrcx) || 0;
	ddata.imagesrcy =		(param && param.imagesrcy) || 0;
	ddata.imagesrcw =		(param && param.imagesrcw) || 0;
	ddata.imagesrch =		(param && param.imagesrch) || 0;
	return ddata;
}

UI_Drawdata_Image.prototype = {
	draw:	function(ui) {
		this.drawbackground(ui);
		var ctx = ui.ctx;
		var style = (this.on) ? ((this.mouseover) ? this.moverstyle : this.style) : ((this.mouseover) ? this.moveroffstyle : this.offstyle);
		var x = this.x + style.lineWidth + style.px;
		var y = this.y + style.lineWidth + style.py;
		if (style.align === "center") {
			x = this.x + (this.width - this.imagewidth) / 2;
		}
		else if (style.align === "right") {
			x = this.x + this.width -  this.imagewidth - style.lineWidth - style.px;
		}
		if (style.valign === "middle") {
			y = this.y + (this.height - this.imageheight) / 2;
		}
		else if (style.valign === "bottom") {
			y = this.y + this.height - this.imageheight - style.lineWidth - style.py;
		}
		if (this.transform) {
			ctx.save();
			if (this.transform === "reversex") {
				ctx.scale(-1, 1);
				ctx.translate(-x * 2 - this.imagewidth, 0);
			}
		}
		if (this.imagesrcw) {
			ctx.drawImage(this.image, this.imagesrcx, this.imagesrcy, this.imagesrcw, this.imagesrch, x, y, this.imagewidth, this.imageheight);
		}
		else {
			ctx.drawImage(this.image, x, y, this.imagewidth, this.imageheight);
		}
		if (this.transform) {
			ctx.restore();
		}
	}
};

Object.setPrototypeOf(UI_Drawdata_Image.prototype, UI_Drawdata.prototype);





function UI2(type, param) {
	var ui = Object.create(UI2.prototype);
	ui.type = type;
	ui.id =							(param && param.id !== undefined) ? param.id : ui.type;
	ui.x =							(param && param.x) || 0;
	ui.y = 							(param && param.y) || 0;
	ui.dx =							(param && param.dx) || 0;
	ui.dy = 						(param && param.dy) || 0;
	ui.px =							(param && param.px) || 0;
	ui.py = 						(param && param.py) || 0;
	ui.param = 					(param && param.param) || null;
	ui.position = 			(param && param.position) || "auto";
	ui.width	=					(param && param.width) || 0;
	ui.height	=					(param && param.height) || 0;
	ui.autosize =				(param && param.autosize) || true;
	ui.style =					(param && param.style) ? UI_Style(param.style) : null;
	ui.moverstyle =			(param && param.moverstyle) ? UI_Style(param.moverstyle) : ui.style;
	ui.offstyle =				(param && param.offstyle) ? UI_Style(param.offstyle) : ui.style;
	ui.moveroffstyle =	(param && param.moveroffstyle) ? UI_Style(param.moveroffstyle) : ui.offstyle;
	ui.arg =						(param && param.arg) || null;
	ui.canvas =					(param && param.canvas) || null;
	ui.ctx =						(param && param.ctx) || null;
	ui.data	=						(param && param.data) || [];
	ui.display =				(param && param.display) || true;
	ui.enablewheel =		(param && param.enablewheel) || false;
	if (param && param.preprocess) {
		ui.preprocess = param.preprocess;
	}
	if (param && param.bindvar) {
		ui.bindvar = param.bindvar;
	}
	if (param && param.onchange) {
		ui.onchange = param.onchange;
	}

	if (ui.data.forEach === undefined) {
		ui.data = [ ui.data ];
	}
	ui.mouseover =	false;
	return ui;
}

UI2.eventlist = [
	"onmousedown",
	"onmouseup",
	"onmousemove",
	"onwheel",
	"onmouseout",
];

UI2.prototype = {
	update:				function() {
		var arg =	this.arg;
		var x =		this.x;
		var y =		this.y;
		this.rect =		UI_Rect(x, y, this.width, this.height);
		this.data.forEach(function(data, i, a) {
			data.x =		x;
			data.y =		y;
			data.arg =	arg;
			data.update();
		});
	},
	drawcommon:	function(th) {
		if (!this.display) {
			return;
		}
		this.preprocessorig();
		this.preprocess();
		var ctx = th.ctx;
		var style = (this.on) ? ((this.mouseover) ? this.moverstyle : this.style) : ((this.mouseover) ? this.moveroffstyle : this.offstyle);
		if (style !== null) {
			ctx.beginPath();
			ctx.rect(this.x, this.y, this.width, this.height);
			ctx.fillStyle = style.fillStyle;
			ctx.fill();
			if (style.lineWidth > 0) {
				ctx.beginPath();
				ctx.lineWidth = style.lineWidth;
				ctx.strokeStyle = style.strokeStyle;
				ctx.rect(this.x + style.lineWidth / 2, this.y + style.lineWidth / 2, this.width - style.lineWidth, this.height - style.lineWidth);
				ctx.strokeStyle = style.strokeStyle;
				ctx.stroke();
			}
		}
		this.draw(th);
	},
	updatebindvar : function() {
	},
	preprocessorig: function() {
	},
	preprocess: function() {
	},
	onchangeorig: function() {
	},
	onchange: function() {
	},
	draw: function(th) {
		this.data[0].draw(th);
	},
};

function UI_Container(param) {
	var container = Object.create(UI_Container.prototype);
	Object.assign(container, UI2("container", param));
	container.space =					(param && param.space) || 0;
	container.lineheight =		(param && param.lineheight) || 0;
	container.update();
	return container;
}

UI_Container.prototype = {
	update:	function() {
		var ctx = this.ctx;
		var canvas = this.canvas;
		var arg = this.arg;
		var t = this;
		var x = this.x + this.px;
		var y = this.y + this.py;
		this.maxx = 0;
		this.maxy = 0;
		this.data.forEach(function(data, i, a) {
			if (data.position === "absolute") {
			}
			if (data.position === "cabsolute") {
				x = t.x + data.dx;
				y = t.y + data.dy;
				data.x = x;
				data.y = y;
			}
			else if (data.position === "relative") {
				x += data.dx;
				y += data.dy;
				data.x = x;
				data.y = y;
			}
			else if (data.position === "newline") {
				x = t.x + t.px + data.dx;
				y += data.dy + t.lineheight;
				data.x = x;
				data.y = y;
			}
			else {
				data.x =		x;
				data.y =		y;
			}
			if (t.maxx < data.x + data.width) {
				t.maxx = data.x + data.width;
			}
			if (t.maxy < data.y + data.height) {
				t.maxy = data.y + data.height;
			}
			data.ctx =	ctx;
			data.canvas = canvas;
			data.arg		= arg;
			data.update();
			x = data.x + data.width + t.space;
		});
		if (this.autosize) {
			this.width = this.maxx - this.x;
			this.height = this.maxy - this.y;
		}
		this.rect =		UI_Rect(this.x, this.y, this.width, this.height);
	},
	draw: function() {
		var bindvar = this.bindvar;
		var bindvarvalue = bindvar && eval(bindvar);
		var arg = this.arg;
		this.data.forEach(function(data, i, a) {
			if (bindvar) {
				data.display = bindvarvalue === data.id;
			}
			data.drawcommon(arg);
		});
	},
};

Object.setPrototypeOf(UI_Container.prototype, UI2.prototype);

function UI_Base(param) {
	var container = Object.create(UI_Base.prototype);
	Object.assign(container, UI_Container(param));
	container.type = "base";
	if (container.canvas) {
		var canvas = container.canvas;
		canvas.ui = container;
		container.eventorig = {};
		UI2.eventlist.forEach(function(data, i, a) {
			container.eventorig[data] = canvas[data];
			canvas[data] = container[data + "prev"];
		});
	}
	container.update();
	return container;
}

UI_Base.prototype = {
	draw: function() {
		var bindvar = this.bindvar;
		var bindvarvalue = bindvar && eval(bindvar);
		var arg = this.arg;
		this.data.forEach(function(data, i, a) {
			if (bindvar) {
				data.display = bindvarvalue === data.id;
			}
			data.drawcommon(arg);
		});
	},
	paint: function() {
		this.drawcommon(this);
	}
};

UI2.eventlist.forEach(function(data, i, a) {
	UI_Base.prototype[data + "prev"] = function(evt) {
		if (this.ui.eventorig[data]) {
			this.ui.eventorig[data].apply(this.ui.canvas, [ evt ] );
		}
		this.ui[data + "mouse"](evt);
		this.ui[data + "own"](evt);
		this.ui[data].apply(this.ui, [ evt ]);
		this.ui.drawcommon(this.ui);
	};

	UI_Base.prototype[data + "mouse"] = function(evt) {
		if (!this.display) {
			return;
		}
		var arg = this.arg;
		var brect = this.canvas.getBoundingClientRect();
		evt.mx = Math.floor(evt.clientX - brect.left);
		evt.my = Math.floor(evt.clientY - brect.top);
		this.mouseover = this.rect.isinside(evt.mx, evt.my);

		this.data.forEach(function(data2, i, a) {
			data2[data + "mouse"](evt);
			if (data === "onwheel" || (data2.mouseover = data2.rect.isinside(evt.mx, evt.my))) {
				data2[data + "own"](evt);
				data2[data].apply(arg, [ evt ]);
			}
		});
	};
	UI_Base.prototype[data + "own"] = function(evt) {
	}

	UI2.prototype[data + "mouse"] = function(evt) {
		if (!this.display) {
			return;
		}
		var arg = this.arg;
		var type = this.type;

		this.data.forEach(function(data2, i, a) {
			if (type === "container") {
				data2[data + "mouse"](evt);
			}
			data2.mouseover = data2.rect.isinside(evt.mx, evt.my);
			if (type === "container" && (data === "onwheel" || data2.mouseover)) {
				data2[data + "own"](evt);
				data2[data].apply(arg, [ evt ]);
			}
		});
	};
	UI2.prototype[data + "own"] = function(evt) {
	};
	UI2.prototype[data] = function(evt) {
	};
});

Object.setPrototypeOf(UI_Base.prototype, UI_Container.prototype);

function UI_Button(param) {
	var button = Object.create(UI_Button.prototype);
	Object.assign(button, UI2("button", param));
	if (!param || !param.moveroffstyle) {
		button.moveroffstyle = button.moverstyle;
	}
	button.update();
	return button;
}

UI_Button.prototype = {
	update:	function() {
		if (this.data.length > 0) {
			this.data[0].arg =	this.arg;
			this.data[0].x =		this.x + this.px;
			this.data[0].y =		this.y + this.py;
			this.data[0].update();
			if (this.autosize) {
				this.width = this.px * 2 + this.data[0].width;
				this.height = this.py * 2 + this.data[0].height;
			}
		}
		this.rect =		UI_Rect(this.x, this.y, this.width, this.height);
	},
};

Object.setPrototypeOf(UI_Button.prototype, UI2.prototype);

function UI_Selector(param) {
	var selector = Object.create(UI_Selector.prototype);
	Object.assign(selector, UI2("selector", param));
	selector.selectnum =		(param && param.selectnum) || 0;
	selector.space =				(param && param.space) || 0;
	selector.lineheight =		(param && param.lineheight) || 10;
	selector.selectid = "none";
	selector.update();
	return selector;
}

UI_Selector.prototype = {
	update:	function() {
		var t = this;
		var arg =	this.arg;
		var x =		this.x + this.px;
		var y =		this.y + this.py;
		var selectnum = +this.selectnum;
		if (selectnum < 0 && selectnum >= this.data.length) {
			selectnum = this.selectnum = 0;
		}
		if (this.data.length > 0) {
			this.selectid = this.data[selectnum].id;
		}
		else {
			this.selectid = "";
		}
	//	this.updatebindvar();
		this.maxx = 0;
		this.maxy = 0;
		this.data.forEach(function(data, i, a) {
			data.x =		x;
			data.y =		y;
			data.arg =	arg;
			if (t.maxx < x + data.width) {
				t.maxx = x + data.width;
			}
			if (t.maxy < y + data.height) {
				t.maxy = y + data.height;
			}
			data.update();
			data.on = (i == selectnum);
			x += data.width + t.space;
			if (data.position === "relative") {
				x += data.dx;
				y += data.dy;
			}
			else if (data.position === "newline") {
				x = data.dx + t.x + t.px;
				y += t.lineheight + data.dy;
			}
		});
		if (this.autosize) {
			this.width = this.maxx - this.x;
			this.height = this.maxy - this.y;
		}
		this.rect =		UI_Rect(this.x, this.y, this.width, this.height);
	},
	preprocessorig: function() {
		if (this.arg && this.bindvar) {
			var selectid = eval(this.bindvar);
			if (this.selectid !== selectid) {
				var selectnum = -1;
				this.data.forEach(function(data, i, a) {
					if (data.id === selectid) {
						selectnum = i;
					}
				});
				if (selectnum >= 0) {
					this.selectid = selectid;
					this.selectnum = selectnum;
					this.onchangeorig();
					this.onchange();
				}
				else {
					this.updatebindvar();
				}
			}
		}
	},
	updatebindvar: function() {
		if (this.arg && this.bindvar) {
			eval(this.bindvar + " = this.selectid;");
		}
	},
	draw:	function(ui) {
		let selectnum = this.selectnum;
		this.data.forEach(function(data, i, a) {
			data.on = (selectnum == i);
			data.draw(ui);
		});
	},
	onmousedownown: function(evt) {
		var newselectnum = -1;
		this.data.forEach(function(data, i, a) {
			if (data.mouseover) {
				newselectnum = i;
			}
		});
		if (newselectnum != -1) {
			this.selectnum = newselectnum;
		}
		this.selectid = this.data[this.selectnum].id;
		this.updatebindvar();
		this.onchangeorig();
		this.onchange();
	},
	onwheelown:	function(evt) {
		if (this.enablewheel && this.data.length > 0) {
			if (evt.deltaY > 0) {
				this.selectnum = (this.selectnum + 1) % this.data.length;
			}
			else if (evt.deltaY < 0) {
				this.selectnum = (this.selectnum + this.data.length - 1) % this.data.length;
			}
			this.selectid = this.data[this.selectnum].id;
			this.updatebindvar();
			this.onchangeorig();
			this.onchange();
		}
	}
};

Object.setPrototypeOf(UI_Selector.prototype, UI2.prototype);

function UI_ClickSelector(param) {
	var selector = Object.create(UI_ClickSelector.prototype);
	Object.assign(selector, UI2("clickselector", param));
	selector.selectnum =		(param && param.selectnum) || 0;
	selector.selectid = "none";
	if (!param || !param.moveroffstyle) {
		selector.moveroffstyle = selector.moverstyle;
	}
	selector.update();
	return selector;
}

UI_ClickSelector.prototype = {
	update:	function() {
		var t = this;
		var arg =	this.arg;
		var x =		this.x + this.px;
		var y =		this.y + this.py;
		var selectnum = +this.selectnum;
		if (selectnum < 0 && selectnum >= this.data.length) {
			selectnum = this.selectnum = 0;
		}
		if (this.data.length > 0) {
			this.selectid = this.data[selectnum].id;
		}
		else {
			this.selectid = "";
		}
	//	this.updatebindvar();
		this.maxwidth = 0;
		this.maxheight = 0;
		this.data.forEach(function(data, i, a) {
			data.x =		x;
			data.y =		y;
			data.arg =	arg;
			if (t.maxwidth < data.width) {
				t.maxwidth = data.width;
			}
			if (t.maxheight < data.height) {
				t.maxheight = data.height;
			}
			data.update();
		});
		if (this.autosize) {
			this.width = this.maxwidth + this.px * 2;
			this.height = this.maxheight + this.py * 2;
		}
		this.rect =		UI_Rect(this.x, this.y, this.width, this.height);
	},
	preprocessorig: function() {
		if (this.arg && this.bindvar) {
			var selectid = eval(this.bindvar);
			if (this.selectid !== selectid) {
				var selectnum = -1;
				this.data.forEach(function(data, i, a) {
					if (data.id === selectid) {
						selectnum = i;
					}
				});
				if (selectnum >= 0) {
					this.selectid = this.data[selectnum].id;
					this.selectnum = selectnum;
					this.onchangeorig();
					this.onchange();
				}
				else {
					eval(this.bindvar + " = this.selectid;");
				}
			}
		}
	},
	updatebindvar: function() {
		if (this.arg && this.bindvar) {
			eval(this.bindvar + " = this.selectid;");
		}
	},
	draw:	function(ui) {
		if (this.data.length > 0) {
			this.data[this.selectnum].draw(ui);
		};
	},
	onmousedownown: function(evt) {
		if (this.data.length > 0) {
			if (evt.buttons & 1) {
				this.selectnum = (this.selectnum + 1) % this.data.length;
			}
			else if (evt.buttons & 2) {
				this.selectnum = (this.selectnum + this.data.length - 1) % this.data.length;
			}
			this.selectid = this.data[this.selectnum].id;
			this.updatebindvar();
			this.onchangeorig();
			this.onchange();
		}
	}
};

Object.setPrototypeOf(UI_ClickSelector.prototype, UI2.prototype);

function UI_TextBox(param) {
	var textbox = Object.create(UI_TextBox.prototype);
	Object.assign(textbox, UI2("textbox", param));
	if (!param || !param.moveroffstyle) {
		textbox.moveroffstyle = textbox.moverstyle;
	}
	textbox.value		= (param && param.value) || 0;
	textbox.msg			= (param && param.msg) || "";
	if (param && param.onchange) {
		textbox.onchange = param.onchange;
	}
	textbox.update();
	return textbox;
}

UI_TextBox.prototype = {
	update:	function() {
		if (this.data.length > 0) {
			this.data[0].arg =	this.arg;
			this.data[0].x =		this.x + this.px;
			this.data[0].y =		this.y + this.py;
			this.data[0].update();
			if (this.autosize) {
				this.width = this.px * 2 + this.data[0].width;
				this.height = this.py * 2 + this.data[0].height;
			}
		}
		this.rect =		UI_Rect(this.x, this.y, this.width, this.height);
	},
	preprocessorig: function() {
		if (this.arg && this.bindvar) {
			eval("this.value = " + this.bindvar + ";");
		}
		if (this.data.length > 0) {
			this.data[0].text = this.value;
		}
	},
	updatebindvar: function() {
		if (this.arg && this.bindvar) {
			eval(this.bindvar + " = this.value;");
		}
	},
	onmousedownown: function() {
		let text = window.prompt(this.msg, this.value);
		if (text !== null) {
			let check = this.onchange(text);
			if (check.result) {
				this.value = check.text;
				this.updatebindvar();
			}
		}
	},
	onchange: function(text) {
		return { result: true, text: text };
	},
};

Object.setPrototypeOf(UI_TextBox.prototype, UI2.prototype);
